package org.jvnet.jaxbcommons.ai.addon;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jvnet.jaxbcommons.addon.AbstractParameterizableCodeAugmenter;
import org.jvnet.jaxbcommons.util.ClassUtils;
import org.jvnet.jaxbcommons.util.CodeModelUtils;
import org.xml.sax.ErrorHandler;

import com.sun.codemodel.JCodeModel;
import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JType;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;

public class AddOn extends AbstractParameterizableCodeAugmenter {

  protected Log logger = LogFactory.getLog(getClass());

  private List interfaces = new ArrayList(10);

  public void setInterface(String name) {
    try {
      final Class interfaceClass = Class.forName(name);
      interfaces.add(interfaceClass);
    }
    catch (ClassNotFoundException ex) {
      logger.error("Bad interface class name.", ex);
    }

  }

  public List getInterfaces() {
    return interfaces;
  }

  public String getOptionName() {
    return "Xautoimplement";
  }

  public String getUsage() {
    return "  -Xautoimplement                :  automactially implements given interfaces";
  }

  protected boolean matches(JCodeModel codeModel, JDefinedClass jinterface, Class theInterface) {
    final Method[] methods = theInterface.getMethods();
    for (int index = 0; index < methods.length; index++) {
      final Method method = methods[index];
      if (!matches(codeModel, jinterface, method))
        return false;
    }
    return true;
  }

  protected boolean matches(JCodeModel codeModel, JDefinedClass jinterface, Method theMethod) {
    for (Iterator itr = jinterface.methods(); itr.hasNext();) {
      final JMethod m = (JMethod) itr.next();
      if (matches(codeModel, m, theMethod))
        return true;
    }
    return false;
  }

  protected boolean matches(JCodeModel codeModel, JMethod m, Method theMethod) {

    if (m.name().equals(theMethod.getName())
        && matches(codeModel, m.type(), theMethod.getReturnType())) {
      final Class[] parameterClasses = theMethod.getParameterTypes();
      final JType[] parameterTypes = m.listParamTypes();
      if (parameterClasses.length == parameterTypes.length) {
        for (int index = 0; index < parameterClasses.length; index++) {
          final Class theClass = parameterClasses[index];
          final JType type = parameterTypes[index];
          if (!matches(codeModel, type, theClass)) {
            return false;
          }
        }
        return true;
      }
    }
    return false;
  }

  protected boolean matches(JCodeModel codeModel, JType type, Class theClass) {
    final JType classType;
    if (Void.TYPE.equals(theClass)) {
      return codeModel.VOID.equals(type);
    }
    else if (theClass.isPrimitive()) {
      Class wrapperClass = org.apache.commons.lang.ClassUtils.primitiveToWrapper(theClass);
      classType = codeModel.ref(wrapperClass).getPrimitiveType();
    }
    else {
      classType = codeModel.ref(theClass);
    }
    return type.equals(classType);
  }

  /**
   * Augments the code.
   * 
   * @param grammar
   *            annotated grammar.
   * @param generatorContext
   *            generator context.
   * @param options
   *            options.
   * @param errorHandler
   *            error handler.
   * @return If the add-on executes successfully, return true. If it detects
   *         some errors but those are reported and recovered gracefully,
   *         return false.
   */
  public boolean run(
      final AnnotatedGrammar grammar,
      final GeneratorContext generatorContext,
      final Options options,
      final ErrorHandler errorHandler) {

    logger.debug("Starting code augmentation.");

    // Process class items
    final ClassItem[] classItems = grammar.getClasses();
    for (int index = 0; index < classItems.length; index++) {
      final ClassItem classItem = classItems[index];
      final ClassContext classContext = generatorContext.getClassContext(classItem);
      logger.debug("Processing [" + classItem.name + "].");
      process(classContext);
    }
    logger.debug("Finished code augmentation.");
    return true;
  }

  protected void process(ClassContext classContext) {

    final JCodeModel codeModel = CodeModelUtils.getCodeModel(classContext);
    final JDefinedClass jnterface = classContext.ref;

    for (Iterator iterator = getInterfaces().iterator(); iterator.hasNext();) {
      final Class theInterface = (Class) iterator.next();
      if (matches(codeModel, jnterface, theInterface)) {
        logger.debug("Interface ["
            + classContext.ref.name()
            + "] will extend the interface ["
            + theInterface.getName()
            + "].");
        ClassUtils._implements(classContext.ref, codeModel.ref(theInterface));
      }
    }
  }
}
