package org.jvnet.jaxbcommons.cloneable.addon.generator;

import java.util.Collection;
import java.util.Iterator;

import org.jvnet.jaxbcommons.util.AccessorUtils;
import org.jvnet.jaxbcommons.util.TypeUtils;
import org.jvnet.jaxbcommons.visitor.ClassifyingVisitor;

import com.sun.codemodel.JBlock;
import com.sun.codemodel.JExpr;
import com.sun.codemodel.JExpression;
import com.sun.codemodel.JForLoop;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JMod;
import com.sun.codemodel.JPrimitiveType;
import com.sun.codemodel.JStatement;
import com.sun.codemodel.JType;
import com.sun.codemodel.JVar;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.TypeItem;

public class FieldCopyVisitor extends ClassifyingVisitor {

  private final JExpression cloned;
  private final JExpression source;

  public FieldCopyVisitor(ClassContext classContext, JExpression cloned, JExpression source) {
    super(classContext);
    this.cloned = cloned;
    this.source = source;
  }

  public Object onSingleField(FieldItem fieldItem) {
    final JMethod getter = AccessorUtils.get(classContext, fieldItem);
    final JMethod setter = AccessorUtils.set(classContext, fieldItem);
    final TypeItem typeItem = TypeUtils.getCommonBaseTypeItem(classContext, fieldItem);
    final JType type = typeItem != null ? typeItem.getType() : getter.type();

    final JBlock block = JBlock.dummyInstance.block();


    final JVar value = block.decl(JMod.FINAL, type, "value", source.invoke(getter));
    final JVar valueCopy = block.decl(type, "valueCopy");

    final JStatement statement = (JStatement) fieldItem.visit(new EntryCopyVisitor(
        classContext,
        value,
        valueCopy, type));
    block.add(statement);

    block.invoke(cloned, setter).arg(valueCopy);

    return block;
  }

  public Object onCollectionField(FieldItem fieldItem) {
    final JBlock block = JBlock.dummyInstance.block();

    final JVar value = block.decl(JMod.FINAL, codeModel.ref(Collection.class), "value", source
        .invoke(AccessorUtils.get(classContext, fieldItem)));

    final JVar valueCopy = block.decl(
        JMod.FINAL,
        codeModel.ref(Collection.class),
        "valueCopy",
        cloned.invoke(AccessorUtils.get(classContext, fieldItem)));

    final TypeItem draftTypeItem = TypeUtils.getCommonBaseTypeItem(classContext, fieldItem);

    final JType draftType = draftTypeItem != null ? draftTypeItem.getType() : codeModel
        .ref(Object.class);

    final JType type = draftType.isPrimitive()
        ? ((JPrimitiveType) draftType).getWrapperClass()
        : draftType;

    final JForLoop forLoop = block._for();

    final JVar iterator = forLoop.init(codeModel.ref(Iterator.class), "iterator", value
        .invoke("iterator"));

    forLoop.test(iterator.invoke("hasNext"));
    final JVar next = forLoop.body().decl(
        JMod.FINAL,
        type,
        "next",
        JExpr.cast(type, iterator.invoke("next")));
    final JVar nextCopy = forLoop.body().decl(type, "nextCopy");

    final JStatement st = (JStatement) fieldItem.visit(new EntryCopyVisitor(
        classContext,
        next,
        nextCopy, type));

    forLoop.body().add(st);
    forLoop.body().invoke(valueCopy, "add").arg(nextCopy);
    return block;
  }
}
