package org.jvnet.jaxbcommons.copyable.addon;

import org.jvnet.jaxbcommons.addon.AbstractParameterizableCodeAugmenter;
import org.jvnet.jaxbcommons.lang.CopyTo;
import org.jvnet.jaxbcommons.lang.builder.CopyBuilder;
import org.jvnet.jaxbcommons.lang.builder.JAXBCopyBuilder;
import org.jvnet.jaxbcommons.util.AccessorUtils;
import org.jvnet.jaxbcommons.util.ClassUtils;
import org.jvnet.jaxbcommons.util.FieldUtils;
import org.xml.sax.ErrorHandler;

import com.sun.codemodel.JBlock;
import com.sun.codemodel.JCodeModel;
import com.sun.codemodel.JConditional;
import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JExpr;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JMod;
import com.sun.codemodel.JOp;
import com.sun.codemodel.JType;
import com.sun.codemodel.JVar;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.FieldUse;

public class AddOn extends AbstractParameterizableCodeAugmenter {

	public String getOptionName() {
		return "Xcopyable";
	}

	public String getUsage() {
		return "TBD";
	}

	private Class copyBuilderClass = JAXBCopyBuilder.class;

	public void setCopyBuilderClass(final Class copyBuilderClass) {
		this.copyBuilderClass = copyBuilderClass;
	}

	public Class getCopyBuilderClass() {
		return copyBuilderClass;
	}

	public boolean run(final AnnotatedGrammar grammar,
			final GeneratorContext generatorContext, final Options options,
			final ErrorHandler errorHandler) {

		logger.debug("Starting code augmentation.");

		// Process class items
		final ClassItem[] classItems = grammar.getClasses();
		logger.debug("Processing [" + classItems.length + "] class items.");
		for (int index = 0; index < classItems.length; index++) {
			final ClassItem classItem = classItems[index];
			final ClassContext classContext = generatorContext
					.getClassContext(classItem);
			logger.debug("Processing [" + classItem.name + "].");
			processClassContext(classContext);
		}
		logger.debug("Finished processing class items.");
		logger.debug("Finished code augmentation.");
		return true;

	}

	protected void processClassContext(ClassContext classContext) {

		final JMethod copyTo$copyTo = generateCopyTo$CopyTo(classContext);
		final JMethod copyable$copyTo = generateCopyable$CopyTo(classContext);
		final JMethod cloneable$clone = generateCloneable$Clone(classContext);
		ClassUtils._implements(classContext.ref, classContext.ref.owner().ref(
				CopyTo.class));

	}

	protected JMethod generateCopyable$CopyTo(final ClassContext classOutline) {
		final JDefinedClass theClass = classOutline.implClass;
		// ClassUtils._implements(theClass,
		// theClass.owner().ref(Copyable.class));
		// ClassUtils._implements(classOutline.ref,
		// theClass.owner().ref(Copyable.class));

		final JMethod copyable$copyTo = theClass.method(JMod.PUBLIC, theClass
				.owner().ref(Object.class), "copyTo");
		{
			final JVar target = copyable$copyTo.param(Object.class, "target");
			final JBlock body = copyable$copyTo.body();

			final JVar copyBuilder = body.decl(JMod.FINAL, theClass.owner()
					.ref(CopyBuilder.class), "copyBuilder", JExpr._new(theClass
					.owner().ref(getCopyBuilderClass())));

			body._return(JExpr.invoke("copyTo").arg(target).arg(copyBuilder));
		}
		return copyable$copyTo;
	}

	protected JMethod generateCloneable$Clone(final ClassContext classOutline) {
		final JDefinedClass theClass = classOutline.implClass;
		// ClassUtils._implements(theClass,
		// theClass.owner().ref(Cloneable.class));
		// ClassUtils._implements(classOutline.ref,
		// theClass.owner().ref(Cloneable.class));

		final JMethod cloneable$Clone = theClass.method(JMod.PUBLIC, theClass
				.owner().ref(Object.class), "clone");
		{
			final JBlock body = cloneable$Clone.body();

			final JVar target = body.decl(classOutline.ref, "target", JExpr
					._new(theClass));

			body._return(JExpr.invoke("copyTo").arg(target));
		}
		return cloneable$Clone;
	}

	protected JMethod generateCopyTo$CopyTo(ClassContext classOutline) {

		final JDefinedClass theClass = classOutline.implClass;
		final JCodeModel codeModel = theClass.owner();

		final JMethod copyTo = theClass.method(JMod.PUBLIC, theClass.owner()
				.ref(Object.class), "copyTo");
		{
			final JVar target = copyTo.param(Object.class, "target");
			final JVar copyBuilder = copyTo.param(CopyBuilder.class,
					"copyBuilder");

			final JBlock body = copyTo.body();

			final JVar draftCopy = body.decl(JMod.FINAL, codeModel
					.ref(Object.class), "draftCopy",

			JOp.cond(JOp.eq(target, JExpr._null()), JExpr._new(theClass),
					target));

			if (classOutline.target.getSuperClass() != null) {
				body.invoke(JExpr._super(), "copyTo").arg(draftCopy).arg(
						copyBuilder);
			}

			final FieldUse[] fieldUses = classOutline.target
					.getDeclaredFieldUses();

			if (fieldUses.length > 0) {
				final JBlock ifInstance = body._if(
						draftCopy._instanceof(classOutline.ref))._then();

				final JVar copy = ifInstance.decl(classOutline.ref, "copy",
						JExpr.cast(classOutline.ref, draftCopy));

				for (int index = 0; index < fieldUses.length; index++) {
					final FieldUse fieldUse = fieldUses[index];
					final FieldItem fieldItem = FieldUtils
							.getFieldItem(fieldUse);

					if (!FieldUtils.isConstantField(classOutline, fieldItem)) {

						final JBlock block = ifInstance.block();

						JMethod get = AccessorUtils
								.get(classOutline, fieldItem);
						JMethod isSet = AccessorUtils.isSet(classOutline,
								fieldItem);
						JMethod unset = AccessorUtils.unset(classOutline,
								fieldItem);
						final JType type = get.type();

						final JVar sourceField = block.decl(type, "source"
								+ fieldUse.name);

						final JBlock assignBlock;
						if (isSet != null && unset != null) {
							final JConditional ifSet = block._if(JExpr._this()
									.invoke(isSet));
							ifSet._else().invoke(copy, unset);
							assignBlock = ifSet._then();
						} else {
							assignBlock = block;
						}

						AccessorUtils.get(assignBlock, sourceField, JExpr
								._this(), classOutline, fieldItem);

						final JVar copyField = assignBlock.decl(type, "copy"
								+ fieldItem.name,

						JExpr.cast(type, JExpr.invoke(copyBuilder, "copy").arg(
								sourceField)));

						AccessorUtils.set(assignBlock, copyField, copy,
								classOutline, fieldItem);
					}
				}
			}
			body._return(draftCopy);
			return copyTo;
		}
	}
}
