/*
 * Created on 28.06.2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package org.jvnet.jaxbcommons.extendedequals.addon.generator;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.Calendar;

import org.jvnet.jaxbcommons.addon.generator.AbstractMethodStrategy;
import org.jvnet.jaxbcommons.lang.EqualsAware;
import org.jvnet.jaxbcommons.lang.VisitedObjects;
import org.jvnet.jaxbcommons.lang.util.EqualsUtils;
import org.jvnet.jaxbcommons.util.AccessorUtils;
import org.jvnet.jaxbcommons.util.ClassUtils;
import org.jvnet.jaxbcommons.util.FieldUtils;

import com.sun.codemodel.JBlock;
import com.sun.codemodel.JClass;
import com.sun.codemodel.JCodeModel;
import com.sun.codemodel.JConditional;
import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JExpr;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JMod;
import com.sun.codemodel.JOp;
import com.sun.codemodel.JVar;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.grammar.ClassItem;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.FieldUse;

/**
 * 
 * @author valikov
 */
public class VisitEqualsStrategy extends AbstractMethodStrategy {

  public JMethod generate(ClassContext classContext, FieldItem fieldItem, JDefinedClass theClass) {

    ClassUtils._implements(classContext.ref, classContext.ref.owner().ref(EqualsAware.class));

    final JMethod visitEquals = classContext.implClass.method(
        JMod.PUBLIC,
        getCodeModel(classContext).BOOLEAN,
        "visitEquals");

    final JVar obj = visitEquals.param(Object.class, "obj");
    final JVar visitedObjects = visitEquals.param(VisitedObjects.class, "visitedObjects");

    visitEquals.body()._if(visitedObjects.invoke("isVisited").arg(JExpr._this()))._then()._return(
        JExpr.TRUE);
    
    visitEquals.body().invoke(visitedObjects, "addVisited").arg(JExpr._this());

    final JConditional ifThisEqObj = visitEquals.body()._if(JOp.eq(JExpr._this(), obj));
    ifThisEqObj._then()._return(JExpr.TRUE);

    final JConditional ifWrongObj = visitEquals.body()._if(
        JOp.cor(JOp.eq(JExpr._null(), obj), JOp.not(JOp._instanceof(obj, classContext.ref))));

    ifWrongObj._then()._return(JExpr.FALSE);
    final JVar target = visitEquals.body().decl(
        classContext.ref,
        "target",
        JExpr.cast(classContext.ref, obj));

    for (ClassItem item = classContext.target; item != null; item = item.getSuperClass()) {
      final ClassContext context = classContext.parent.getClassContext(item);
      generateFieldsEquality(context, visitEquals.body(), target, visitedObjects);
    }

    visitEquals.body()._return(JExpr.TRUE);
    return visitEquals;
  }

  /**
   * Generates field equality check instructions.
   *
   * @param classContext class context.
   * @param methodBody   method body.
   * @param target       target variable.
   */
  protected void generateFieldsEquality(
      final ClassContext classContext,
      final JBlock methodBody,
      final JVar target,
      final JVar visitedObjects) {
    final FieldUse[] fieldUses = classContext.target.getDeclaredFieldUses();
    // Iterate over field uses
    for (int index = 0; index < fieldUses.length; index++) {
      final FieldUse fieldUse = fieldUses[index];
      final FieldItem fieldItem = FieldUtils.getFieldItem(fieldUse);
      final JMethod getter = AccessorUtils.get(classContext, fieldItem);
      if (getter != null) {
        final JBlock block = methodBody.block();
        final JVar value = block.decl(getter.type(), "value", JExpr._this().invoke(getter));
        final JVar targetValue = block.decl(getter.type(), "targetValue", target.invoke(getter));
        valuesEqual(getCodeModel(classContext), block, value, targetValue, visitedObjects);
      }
    }
  }

  /**
   * Generates expression checking equality of two values.
   *
   * @param codeModel code model.
   * @param value value.
   * @param targetValue target value.
   */
  protected void valuesEqual(
      JCodeModel codeModel,
      final JBlock block,
      final JVar value,
      final JVar targetValue,
      final JVar visitedObjects) {

    if (value.type().isArray()) {
      final JConditional ifValuesNotEqual = block._if(JOp.not(codeModel
          .ref(Arrays.class)
          .staticInvoke("equals")
          .arg(value)
          .arg(targetValue)));
      ifValuesNotEqual._then()._return(JExpr.FALSE);
    }
    else if (value.type().isPrimitive()) {
      final JConditional ifValuesNotEqual = block._if(JOp.ne(value, targetValue));
      ifValuesNotEqual._then()._return(JExpr.FALSE);
    }
    else if (value.type() == codeModel.ref(Calendar.class)) {
      final JConditional ifValuesNotEqual = block._if(JOp.not(JOp.cor(
          JOp.eq(value, targetValue),
          JOp.cand(JOp.ne(value, JExpr._null()),

          JOp.eq(value.invoke("getTime").invoke("getTime"), targetValue.invoke("getTime").invoke(
              "getTime"))))));
      ifValuesNotEqual._then()._return(JExpr.FALSE);
    }
    else if (value.type() == codeModel.ref(BigDecimal.class)) {
      final JConditional ifValuesNotEqual = block._if(JOp.not(JOp.cor(
          JOp.eq(value, targetValue),
          JOp.cand(JOp.ne(value, JExpr._null()),

          JOp.eq(value.invoke("compareTo").arg(targetValue), JExpr.lit(0))))));
      ifValuesNotEqual._then()._return(JExpr.FALSE);
    }
    else {
      final JClass equalsUtils = codeModel.ref(EqualsUtils.class);
      block._if(
          JOp.not(equalsUtils
              .staticInvoke("equals")
              .arg(value)
              .arg(targetValue)
              .arg(visitedObjects)))._then()._return(JExpr.FALSE);
    }
  }

}
