package org.jvnet.jaxbcommons.i18n.addon;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jvnet.jaxbcommons.addon.AbstractParameterizableCodeAugmenter;
import org.jvnet.jaxbcommons.enums.addon.util.EnumUtils;
import org.jvnet.jaxbcommons.i18n.addon.util.PropertyUtils;
import org.jvnet.jaxbcommons.util.FieldUtils;
import org.xml.sax.ErrorHandler;

import com.sun.codemodel.JDefinedClass;
import com.sun.msv.datatype.xsd.XSDatatype;
import com.sun.msv.grammar.ValueExp;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.FieldUse;

/**
 * This add-on generates property files base on class, field and enum names.
 * 
 * @author valikov
 */
public class AddOn extends AbstractParameterizableCodeAugmenter {

	/** Logger. */
	protected Log logger = LogFactory.getLog(getClass());

	/** Option name. */
	public static final String OPTION_NAME = "Xi18n";

	private Map importedProperties = new HashMap();

	private Map lowercaseImportedProperties = new HashMap();

	private Map exportedProperties = new HashMap();

	private Map lowercaseExportedProperties = new HashMap();

	// Configurable properties

	private List locales = new ArrayList(3);

	public void setLocale(String locale) {
		locales.add(parseLocale(locale));
	}

	private File importDirectory;

	public void setImportDirectory(File importDirectory) {
		this.importDirectory = importDirectory;
	}

	private String importName = "messages";

	public void setImportName(String importName) {
		this.importName = importName;
	}

	private File exportDirectory;

	public void setExportDirectory(File exportDirectory) {
		this.exportDirectory = exportDirectory;
	}

	private String exportName = "messages";

	public void setExportName(String exportName) {
		this.exportName = exportName;
	}

	public String getOptionName() {
		return OPTION_NAME;
	}

	public String getUsage() {
		return "";
	}

	/**
	 * Parses locale code given as string.
	 * 
	 * @param string
	 *            locale code.
	 * @return Parsed locale.
	 */
	private Locale parseLocale(String string) {
		final StringTokenizer tokens = new StringTokenizer(string, "_");
		final String language = tokens.hasMoreTokens() ? tokens.nextToken()
				: "";
		final String country = tokens.hasMoreTokens() ? tokens.nextToken() : "";
		String variant = "";
		String sep = "";
		while (tokens.hasMoreTokens()) {
			variant += sep + tokens.nextToken();
			sep = "_";
		}
		return new Locale(language, country, variant);
	}

	public boolean run(AnnotatedGrammar grammar, GeneratorContext context,
			Options options, ErrorHandler errorHandler) {

		loadProperties();

		processProperties(grammar, context);

		processLowercaseProperties();

		saveProperties(options);
		return true;
	}

	protected void saveProperties(Options options) {
		for (final Iterator iterator = locales.iterator(); iterator.hasNext();) {
			final Locale locale = (Locale) iterator.next();
			final String l = locale.toString();
			final Map exp = (Map) exportedProperties.get(l);
			final File exportDirectory = this.exportDirectory == null ? options.targetDir
					: this.exportDirectory;

			if (exportDirectory != null && exportName != null) {

				if (!exportDirectory.exists()) {
					logger
							.info(MessageFormat
									.format(
											"Export directory [{0}] does not exist. Creating.",
											new Object[] { exportDirectory
													.getAbsolutePath() }));
					exportDirectory.mkdirs();
				}

				if (exportDirectory.isDirectory()) {
					final String name = exportName + "_" + l + ".properties";
					final File file = new File(exportDirectory, name);
					try {
						// exp.store(new FileOutputStream(file), null);
						PropertyUtils.store(exp, new FileOutputStream(file),
								null);
					} catch (IOException ioex) {
						logger.error("Error storing properties.", ioex);
					}
				} else {
					logger
							.error(MessageFormat
									.format(
											"Export directory [{0}] is not a directory. Unable to export",
											new Object[] { exportDirectory
													.getAbsolutePath() }));
				}
			}
		}
	}

	protected void processLowercaseProperties() {
		for (final Iterator iterator = locales.iterator(); iterator.hasNext();) {
			final Locale locale = (Locale) iterator.next();
			final String l = locale.toString();
			final Map imp = (Map) importedProperties.get(l);
			final Map exp = (Map) exportedProperties.get(l);
			final Map lexp = (Map) lowercaseExportedProperties.get(l);

			for (final Iterator names = imp.keySet().iterator(); names
					.hasNext();) {
				final String key = (String) names.next();
				if (!lexp.containsKey(key.toLowerCase(locale))) {
					final String value = (String) imp.get(key);
					logger.debug(MessageFormat.format("{0}={1}[{2}]",
							new Object[] { key, value, l }));
					exp.put(key, value);
					lexp.put(key.toLowerCase(locale), value);
				}
			}
		}
	}

	protected void processProperties(AnnotatedGrammar grammar,
			GeneratorContext context) {
		final ClassItem[] classes = grammar.getClasses();
		for (int index = 0; index < classes.length; index++) {
			final ClassItem classItem = classes[index];
			final ClassContext classContext = context
					.getClassContext(classItem);
			setProperty(classItem.getTypeAsDefined().fullName(), "@@TODO."
					+ classItem.getTypeAsDefined().name() + "@@");

			final FieldUse[] fields = classItem.getDeclaredFieldUses();
			for (int jndex = 0; jndex < fields.length; jndex++) {
				final FieldUse fieldUse = fields[jndex];
				setProperty(fieldUse.owner.getType().fullName() + "."
						+ fieldUse.name, "@@TODO." + fieldUse.name + "@@");

				if (fieldUse.type instanceof JDefinedClass
						&& null == grammar
								.getClassItem((JDefinedClass) fieldUse.type)) {

					final JDefinedClass enumClass = (JDefinedClass) fieldUse.type;
					final FieldItem fieldItem = FieldUtils
							.getFieldItem(fieldUse);

					final Collection enumValues = EnumUtils.getValues(
							classContext, fieldItem);

					// TODO: biconversion
					if (enumValues != null) {
						for (Iterator iterator = enumValues.iterator(); iterator
								.hasNext();) {
							final ValueExp enumValue = (ValueExp) iterator
									.next();

							final String lexical;
							if (enumValue.dt instanceof XSDatatype)
								lexical = ((XSDatatype) enumValue.dt)
										.convertToLexicalValue(enumValue.value,
												null);
							else
								lexical = enumValue.value.toString();
							setProperty(enumClass.fullName() + "." + lexical,
									"@@TODO." + lexical + "@@");

						}
					}
				}
			}
		}
	}

	protected void loadProperties() {
		for (final Iterator iterator = locales.iterator(); iterator.hasNext();) {
			final Locale locale = (Locale) iterator.next();
			loadProperties(locale);
		}
	}

	protected void loadProperties(final Locale locale) {
		final String l = locale.toString();

		if (importDirectory != null && importName != null) {
			final String name = importName + "_" + l + ".properties";
			final File file = new File(importDirectory, name);
			if (file.isFile()) {

				logger.info(MessageFormat.format(
						"Loading properties from [{0}].", new Object[] { file
								.getAbsolutePath() }));
				try {
					final Map properties = PropertyUtils
							.load(new FileInputStream(file));

					final Map lowercaseProperties = PropertyUtils.toLowerCase(
							properties, locale);
					importedProperties.put(l, properties);
					lowercaseImportedProperties.put(l, lowercaseProperties);
				} catch (IOException ioex) {
					logger.error("Error loading properties.", ioex);
					importedProperties.put(l, new TreeMap());
					lowercaseImportedProperties.put(l, new TreeMap());
				}
			} else {
				logger.warn(MessageFormat.format(
						"[{0}] is not a file. Ignoring.", new Object[] { file
								.getAbsolutePath() }));
				importedProperties.put(l, new TreeMap());
				lowercaseImportedProperties.put(l, new TreeMap());
			}
		}
		exportedProperties.put(l, new TreeMap());
		lowercaseExportedProperties.put(l, new TreeMap());
	}

	protected void setProperty(String name, String value) {
		for (final Iterator iterator = locales.iterator(); iterator.hasNext();) {
			final Locale locale = (Locale) iterator.next();
			setProperty(name, value, locale);
		}
	}

	protected void setProperty(String name, String draft, Locale locale) {
		final String l = locale.toString();
		final Map imp = (Map) lowercaseImportedProperties.get(l);
		final Map exp = (Map) exportedProperties.get(l);
		final Map lexp = (Map) lowercaseExportedProperties.get(l);
		final String key = name.toLowerCase(locale);
		final String value = imp.containsKey(key) ? (String) imp.get(key)
				: draft;
		logger.debug(MessageFormat.format("{0}={1}[{2}]", new Object[] { name,
				value, l }));
		exp.put(name, value);
		lexp.put(name.toLowerCase(), value);
	}

}
