package org.jvnet.jaxbcommons.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.grammar.ClassItem;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.FieldUse;
import com.sun.tools.xjc.grammar.JavaItem;
import com.sun.tools.xjc.reader.xmlschema.bindinfo.BIDeclaration;
import com.sun.tools.xjc.reader.xmlschema.bindinfo.BIXPluginCustomization;

public class CustomizationUtils {

  private static final DocumentBuilder DOCUMENT_BUILDER;
  static {
    final DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
    try {
      DOCUMENT_BUILDER = documentBuilderFactory.newDocumentBuilder();
    }
    catch (ParserConfigurationException pce) {
      throw new ExceptionInInitializerError(pce);
    }
  }

  public static Collection getCustomizations(JavaItem item) {
    final Collection customizations = new ArrayList();
    for (Iterator iterator = item.declarations.iterator(); iterator.hasNext();) {
      final BIDeclaration declaration = (BIDeclaration) iterator.next();
      if (declaration instanceof BIXPluginCustomization) {
        customizations.add(declaration);
      }
    }
    return customizations;
  }

  public static BIXPluginCustomization createCustomization(QName name) {
    final Document document = DOCUMENT_BUILDER.newDocument();
    final Element element = document.createElementNS(name.getNamespaceURI(), name.getLocalPart());
    return createCustomization(element);
  }

  public static BIXPluginCustomization createCustomization(final Element element) {
    final BIXPluginCustomization customization = new BIXPluginCustomization(element, null);
    return customization;
  }

  public static BIXPluginCustomization getOrCreateCustomization(JavaItem item, QName name) {
    if (containsCustomization(item, name)) {
      return getCustomization(item, name);
    }
    else {
      final BIXPluginCustomization customization = createCustomization(name);
      item.declarations.add(customization);
      return customization;
    }
  }

  public static BIXPluginCustomization[] getCustomizations(JavaItem item, QName name) {
    final Collection customizationsDraft = getCustomizations(item);
    final Collection customizations = new ArrayList();

    for (Iterator iterator = customizationsDraft.iterator(); iterator.hasNext();) {
      final BIXPluginCustomization customization = (BIXPluginCustomization) iterator.next();

      final Element element = customization.element;
      if (null != element) {
        final QName elementName = new QName(element.getNamespaceURI(), element.getLocalName());
        if (elementName.equals(name))
          customizations.add(customization);
      }
    }
    return (BIXPluginCustomization[]) customizations
        .toArray(new BIXPluginCustomization[customizations.size()]);
  }

  public static boolean containsCustomization(JavaItem item, QName name) {
    for (Iterator iterator = item.declarations.iterator(); iterator.hasNext();) {
      final BIDeclaration declaration = (BIDeclaration) iterator.next();
      if (declaration instanceof BIXPluginCustomization) {
        final BIXPluginCustomization customization = (BIXPluginCustomization) declaration;
        if (name.equals(customization.getName())) {
          return true;
        }
      }
    }
    return false;
  }
  
  public static boolean ancestorOrSelfContainsCustomization(ClassItem item, QName name) {
	  return containsCustomization(item, name) || (item.getSuperClass() ==  null ? false : ancestorOrSelfContainsCustomization(item.getSuperClass(), name));
  }
  
  public static BIXPluginCustomization ancestorOrSelfCustomization(ClassItem item, QName name)
  {
	  if (containsCustomization(item, name))
	  {
		  return getCustomization(item, name);
	  }
	  else if (item.getSuperClass() != null)
	  {
		  return ancestorOrSelfCustomization(item.getSuperClass(), name);
	  }
	  else
	  {
		  return null;
	  }
  }

  public static BIXPluginCustomization getCustomization(JavaItem item, QName name) {
    final BIXPluginCustomization[] customizations = getCustomizations(item, name);
    if (customizations == null || customizations.length < 1) {
      return null;
    }
    else {
      return customizations[0];
    }
  }

  public static FieldItem getFieldItemWithCustomization(ClassContext classContext, final QName name) {
    final FieldUse[] fieldUses = classContext.target.getDeclaredFieldUses();
    for (int index = 0; index < fieldUses.length; index++) {
      final FieldUse fieldUse = fieldUses[index];
      final FieldItem draftFieldItem = FieldUtils.getFieldItem(fieldUse);

      if (containsCustomization(draftFieldItem, name)) {
        return draftFieldItem;
      }
    }
    return null;
  }

  public static FieldItem getInheritedFieldWithCustomization(
      ClassContext classContext,
      final QName name) {
    final FieldUse[] fieldUses = classContext.target.getDeclaredFieldUses();
    for (int index = 0; index < fieldUses.length; index++) {
      final FieldUse fieldUse = fieldUses[index];
      final FieldItem draftFieldItem = FieldUtils.getFieldItem(fieldUse);

      if (containsCustomization(draftFieldItem, name)) {
        return draftFieldItem;
      }
    }
    final ClassItem superClassItem = classContext.target.getSuperClass();
    if (superClassItem != null) {
      final ClassContext superClassContext = classContext.parent.getClassContext(superClassItem);
      return getInheritedFieldWithCustomization(superClassContext, name);
    }
    else {
      return null;
    }
  }

}
